#include <stdio.h>
#include <string.h>
#include <unistd.h>

#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/Xatom.h>
#include <X11/cursorfont.h>

int
ignoreerrors(Display *disp, XErrorEvent *eev)
{
	return 0;
}

static Window
findtarget(Display *disp, Window root)
{
	Cursor cursor;
	XEvent ev;

	cursor = XCreateFontCursor(disp, XC_target);
	XGrabPointer(disp, root, False, ButtonPressMask|ButtonReleaseMask,
	    GrabModeSync, GrabModeAsync, root, cursor, CurrentTime);
	XAllowEvents(disp, SyncPointer, CurrentTime);
	XWindowEvent(disp, root, ButtonPressMask, &ev);
	XUngrabPointer(disp, CurrentTime);
	XFreeCursor(disp, cursor);
	XFlush(disp);

	return ev.xbutton.subwindow;
}

static void
writestring(Display *disp, Window root, Window target, char *string, size_t len)
{
	XKeyEvent kev;
	KeySym keysym;
	KeyCode keycode;
	char *p, *end;

	end = string + len;

	memset(&kev, 0, sizeof(kev));
	kev.display = disp;
	kev.window = target;
	kev.root = root;
	kev.subwindow = None;
	kev.time = CurrentTime;
	kev.same_screen = True;

	for (p = string; p < end; p++) {
		unsigned char c = *p;
		if (c < 0x20) {
			if (c == '\n')
				c = '\r';
			keysym = 0xff00 | c;
		} else if (c >= 0x20 && c <= 0x7f) {
			keysym = c;
		} else {
			keysym = 0;
		}
		if (keysym == 0)
			continue;

		keycode = XKeysymToKeycode(disp, keysym);

		if (keycode == 0)
			continue;

		kev.keycode = keycode;
		kev.type = KeyPress;
		XSendEvent(disp, target, True, KeyPressMask, (XEvent *)&kev);
		kev.type = KeyRelease;
		XSendEvent(disp, target, True, KeyReleaseMask, (XEvent *)&kev);
		XFlush(disp);
	}
}

int
main(int argc, char **argv)
{
	Display *disp;
	int scr;
	Window root, target;
	char buf[256];
	size_t len;

	disp = XOpenDisplay("");

	scr = DefaultScreen(disp);
	root = RootWindow(disp, scr);

	XSetErrorHandler(ignoreerrors);

	target = findtarget(disp, root);

	while (1) {
		len = read(0, buf, sizeof(buf));
		if (len == 0 || len == -1)
			break;
		writestring(disp, root, target, buf, len);
	}

	return 0;
}
